@@ -19,12 +19,7 @@
 // OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 // THE SOFTWARE.
 
-#ifndef CAPNP_COMPILER_NODE_TRANSLATOR_H_
-#define CAPNP_COMPILER_NODE_TRANSLATOR_H_
-
-#if defined(__GNUC__) && !defined(CAPNP_HEADER_WARNINGS)
-#pragma GCC system_header
-#endif
+#pragma once
 
 #include <capnp/orphan.h>
 #include <capnp/compiler/grammar.capnp.h>
@@ -33,6 +28,9 @@
 #include <kj/vector.h>
 #include <kj/one-of.h>
 #include "error-reporter.h"
+#include <map>
+
+CAPNP_BEGIN_HEADER
 
 namespace capnp {
 namespace compiler {
@@ -136,6 +134,9 @@ public:
     // Auxiliary nodes that were produced when translating this node and should be loaded along
     // with it.  In particular, structs that contain groups (or named unions) spawn extra nodes
     // representing those, and interfaces spawn struct nodes representing method params/results.
+
+    kj::Array<schema::Node::SourceInfo::Reader> sourceInfo;
+    // The SourceInfo for the node and all aux nodes.
   };
 
   NodeSet getBootstrapNode();
@@ -147,9 +148,12 @@ public:
   // If the final node has already been built, this will actually return the final node (in fact,
   // it's the same node object).
 
-  NodeSet finish();
+  NodeSet finish(Schema selfUnboundBootstrap);
   // Finish translating the node (including filling in all the pieces that are missing from the
   // bootstrap node) and return it.
+  //
+  // `selfUnboundBootstrap` is a Schema build using the Node returned by getBootstrapNode(), and
+  // with generic parameters "unbound", i.e. it was returned by SchemaLoader::getUnbound().
 
   static kj::Maybe<Resolver::ResolveResult> compileDecl(
       uint64_t scopeId, uint scopeParameterCount, Resolver& resolver, ErrorReporter& errorReporter,
@@ -177,17 +181,25 @@ private:
   Orphan<schema::Node> wipNode;
   // The work-in-progress schema node.
 
-  kj::Vector<Orphan<schema::Node>> groups;
+  Orphan<schema::Node::SourceInfo> sourceInfo;
+  // Doc comments and other source info for this node.
+
+  struct AuxNode {
+    Orphan<schema::Node> node;
+    Orphan<schema::Node::SourceInfo> sourceInfo;
+  };
+
+  kj::Vector<AuxNode> groups;
   // If this is a struct node and it contains groups, these are the nodes for those groups,  which
   // must be loaded together with the top-level node.
 
-  kj::Vector<Orphan<schema::Node>> paramStructs;
+  kj::Vector<AuxNode> paramStructs;
   // If this is an interface, these are the auto-generated structs representing params and results.
 
   struct UnfinishedValue {
     Expression::Reader source;
     schema::Type::Reader type;
-    Schema typeScope;
+    kj::Maybe<Schema> typeScope;
     schema::Value::Builder target;
   };
   kj::Vector<UnfinishedValue> unfinishedValues;
@@ -217,7 +229,7 @@ private:
 
     uint64_t scopeId;
     // If zero, then any reference to an implciit param in this context should be compiled to a
-    // `implicitMethodParam` AnyPointer. If non-zero, it should be complied to a `parameter`
+    // `implicitMethodParam` AnyPointer. If non-zero, it should be compiled to a `parameter`
     // AnyPointer.
 
     List<Declaration::BrandParameter>::Reader params;
@@ -248,15 +260,13 @@ private:
 
   void compileBootstrapValue(
       Expression::Reader source, schema::Type::Reader type, schema::Value::Builder target,
-      Schema typeScope = Schema());
+      kj::Maybe<Schema> typeScope = nullptr);
   // Calls compileValue() if this value should be interpreted at bootstrap time.  Otheriwse,
   // adds the value to `unfinishedValues` for later evaluation.
   //
-  // If `type` comes from some other node, `typeScope` is the schema for that node. This is only
-  // really needed for looking up generic parameter bindings, therefore if the type comes from
-  // the node being built, an empty "Schema" (the default) works here because the node being built
-  // is of course being built for all possible bindings and thus none of its generic parameters are
-  // bound.
+  // If `type` comes from some other node, `typeScope` is the schema for that node. Otherwise the
+  // scope of the type expression is assumed to be this node (meaning, in particular, that no
+  // generic type parameters are bound).
 
   void compileValue(Expression::Reader source, schema::Type::Reader type,
                     Schema typeScope, schema::Value::Builder target, bool isBootstrap);
@@ -308,4 +318,4 @@ private:
 }  // namespace compiler
 }  // namespace capnp
 
-#endif  // CAPNP_COMPILER_NODE_TRANSLATOR_H_
+CAPNP_END_HEADER
