@@ -19,12 +19,7 @@
 // OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 // THE SOFTWARE.
 
-#ifndef CAPNP_TEST_UTIL_H_
-#define CAPNP_TEST_UTIL_H_
-
-#if defined(__GNUC__) && !defined(CAPNP_HEADER_WARNINGS)
-#pragma GCC system_header
-#endif
+#pragma once
 
 #include <capnp/test.capnp.h>
 #include <iostream>
@@ -33,8 +28,11 @@
 
 #if !CAPNP_LITE
 #include "dynamic.h"
+#include <kj/io.h>
 #endif  // !CAPNP_LITE
 
+CAPNP_BEGIN_HEADER
+
 // TODO(cleanup): Auto-generate stringification functions for union discriminants.
 namespace capnproto_test {
 namespace capnp {
@@ -223,6 +221,8 @@ class TestCallOrderImpl final: public test::TestCallOrder::Server {
 public:
   kj::Promise<void> getCallSequence(GetCallSequenceContext context) override;
 
+  uint getCount() { return count; }
+
 private:
   uint count = 0;
 };
@@ -275,6 +275,8 @@ public:
 
   kj::Promise<void> getEnormousString(GetEnormousStringContext context) override;
 
+  kj::Promise<void> writeToFd(WriteToFdContext context) override;
+
 private:
   int& callCount;
   int& handleCount;
@@ -304,9 +306,57 @@ private:
   TestInterfaceImpl impl;
 };
 
+class TestFdCap final: public test::TestInterface::Server {
+  // Implementation of TestInterface that wraps a file descriptor.
+
+public:
+  TestFdCap(kj::AutoCloseFd fd): fd(kj::mv(fd)) {}
+
+  kj::Maybe<int> getFd() override { return fd.get(); }
+
+private:
+  kj::AutoCloseFd fd;
+};
+
+class TestStreamingImpl final: public test::TestStreaming::Server {
+public:
+  uint iSum = 0;
+  uint jSum = 0;
+  kj::Maybe<kj::Own<kj::PromiseFulfiller<void>>> fulfiller;
+  bool jShouldThrow = false;
+
+  kj::Promise<void> doStreamI(DoStreamIContext context) override {
+    iSum += context.getParams().getI();
+    auto paf = kj::newPromiseAndFulfiller<void>();
+    fulfiller = kj::mv(paf.fulfiller);
+    return kj::mv(paf.promise);
+  }
+
+  kj::Promise<void> doStreamJ(DoStreamJContext context) override {
+    context.allowCancellation();
+    jSum += context.getParams().getJ();
+
+    if (jShouldThrow) {
+      KJ_FAIL_ASSERT("throw requested") { break; }
+      return kj::READY_NOW;
+    }
+
+    auto paf = kj::newPromiseAndFulfiller<void>();
+    fulfiller = kj::mv(paf.fulfiller);
+    return kj::mv(paf.promise);
+  }
+
+  kj::Promise<void> finishStream(FinishStreamContext context) override {
+    auto results = context.getResults();
+    results.setTotalI(iSum);
+    results.setTotalJ(jSum);
+    return kj::READY_NOW;
+  }
+};
+
 #endif  // !CAPNP_LITE
 
 }  // namespace _ (private)
 }  // namespace capnp
 
-#endif  // TEST_UTIL_H_
+CAPNP_END_HEADER
