@@ -104,20 +104,13 @@
 //   define it explicitly (e.g. -DKJ_DEBUG).  Generally, production builds should NOT use KJ_DEBUG
 //   as it may enable expensive checks that are unlikely to fail.
 
-#ifndef KJ_DEBUG_H_
-#define KJ_DEBUG_H_
-
-#if defined(__GNUC__) && !KJ_HEADER_WARNINGS
-#pragma GCC system_header
-#endif
+#pragma once
 
 #include "string.h"
 #include "exception.h"
+#include "windows-sanity.h"  // work-around macro conflict with `ERROR`
 
-#ifdef ERROR
-// This is problematic because windows.h #defines ERROR, which we use in an enum here.
-#error "Make sure to to undefine ERROR (or just #include <kj/windows-sanity.h>) before this file"
-#endif
+KJ_BEGIN_HEADER
 
 namespace kj {
 
@@ -127,7 +120,7 @@ namespace kj {
 //   you to request this behavior with "##__VA_ARGS__".
 // - If __VA_ARGS__ is passed directly as an argument to another macro, it will be treated as a
 //   *single* argument rather than an argument list. This can be worked around by wrapping the
-//   outer macro call in KJ_EXPAND(), which appraently forces __VA_ARGS__ to be expanded before
+//   outer macro call in KJ_EXPAND(), which apparently forces __VA_ARGS__ to be expanded before
 //   the macro is evaluated. I don't understand the C preprocessor.
 // - Using "#__VA_ARGS__" to stringify __VA_ARGS__ expands to zero tokens when __VA_ARGS__ is
 //   empty, rather than expanding to an empty string literal. We can work around by concatenating
@@ -136,8 +129,9 @@ namespace kj {
 #define KJ_EXPAND(X) X
 
 #define KJ_LOG(severity, ...) \
-  if (!::kj::_::Debug::shouldLog(::kj::_::Debug::Severity::severity)) {} else \
-    ::kj::_::Debug::log(__FILE__, __LINE__, ::kj::_::Debug::Severity::severity, \
+  for (bool _kj_shouldLog = ::kj::_::Debug::shouldLog(::kj::LogSeverity::severity); \
+       _kj_shouldLog; _kj_shouldLog = false) \
+    ::kj::_::Debug::log(__FILE__, __LINE__, ::kj::LogSeverity::severity, \
                         "" #__VA_ARGS__, __VA_ARGS__)
 
 #define KJ_DBG(...) KJ_EXPAND(KJ_LOG(DBG, __VA_ARGS__))
@@ -165,27 +159,48 @@ namespace kj {
   for (::kj::_::Debug::Fault f(__FILE__, __LINE__, \
            errorNumber, code, "" #__VA_ARGS__, __VA_ARGS__);; f.fatal())
 
+#if _WIN32 || __CYGWIN__
+
+#define KJ_WIN32(call, ...) \
+  if (auto _kjWin32Result = ::kj::_::Debug::win32Call(call)) {} else \
+    for (::kj::_::Debug::Fault f(__FILE__, __LINE__, \
+             _kjWin32Result, #call, "" #__VA_ARGS__, __VA_ARGS__);; f.fatal())
+
+#define KJ_WINSOCK(call, ...) \
+  if (auto _kjWin32Result = ::kj::_::Debug::winsockCall(call)) {} else \
+    for (::kj::_::Debug::Fault f(__FILE__, __LINE__, \
+             _kjWin32Result, #call, "" #__VA_ARGS__, __VA_ARGS__);; f.fatal())
+
+#define KJ_FAIL_WIN32(code, errorNumber, ...) \
+  for (::kj::_::Debug::Fault f(__FILE__, __LINE__, \
+           ::kj::_::Debug::Win32Result(errorNumber), code, "" #__VA_ARGS__, __VA_ARGS__);; f.fatal())
+
+#endif
+
 #define KJ_UNIMPLEMENTED(...) \
   for (::kj::_::Debug::Fault f(__FILE__, __LINE__, ::kj::Exception::Type::UNIMPLEMENTED, \
                                nullptr, "" #__VA_ARGS__, __VA_ARGS__);; f.fatal())
 
+// TODO(msvc):  MSVC mis-deduces `ContextImpl<decltype(func)>` as `ContextImpl<int>` in some edge
+// cases, such as inside nested lambdas inside member functions. Wrapping the type in
+// `decltype(instance<...>())` helps it deduce the context function's type correctly.
 #define KJ_CONTEXT(...) \
   auto KJ_UNIQUE_NAME(_kjContextFunc) = [&]() -> ::kj::_::Debug::Context::Value { \
         return ::kj::_::Debug::Context::Value(__FILE__, __LINE__, \
             ::kj::_::Debug::makeDescription("" #__VA_ARGS__, __VA_ARGS__)); \
       }; \
-  ::kj::_::Debug::ContextImpl<decltype(KJ_UNIQUE_NAME(_kjContextFunc))> \
+  decltype(::kj::instance<::kj::_::Debug::ContextImpl<decltype(KJ_UNIQUE_NAME(_kjContextFunc))>>()) \
       KJ_UNIQUE_NAME(_kjContext)(KJ_UNIQUE_NAME(_kjContextFunc))
 
 #define KJ_REQUIRE_NONNULL(value, ...) \
-  (*({ \
+  (*[&] { \
     auto _kj_result = ::kj::_::readMaybe(value); \
     if (KJ_UNLIKELY(!_kj_result)) { \
       ::kj::_::Debug::Fault(__FILE__, __LINE__, ::kj::Exception::Type::FAILED, \
                             #value " != nullptr", "" #__VA_ARGS__, __VA_ARGS__).fatal(); \
     } \
-    kj::mv(_kj_result); \
-  }))
+    return _kj_result; \
+  }())
 
 #define KJ_EXCEPTION(type, ...) \
   ::kj::Exception(::kj::Exception::Type::type, __FILE__, __LINE__, \
@@ -194,9 +209,10 @@ namespace kj {
 #else
 
 #define KJ_LOG(severity, ...) \
-  if (!::kj::_::Debug::shouldLog(::kj::_::Debug::Severity::severity)) {} else \
-    ::kj::_::Debug::log(__FILE__, __LINE__, ::kj::_::Debug::Severity::severity, \
-                        #__VA_ARGS__, __VA_ARGS__)
+  for (bool _kj_shouldLog = ::kj::_::Debug::shouldLog(::kj::LogSeverity::severity); \
+       _kj_shouldLog; _kj_shouldLog = false) \
+    ::kj::_::Debug::log(__FILE__, __LINE__, ::kj::LogSeverity::severity, \
+                        #__VA_ARGS__, ##__VA_ARGS__)
 
 #define KJ_DBG(...) KJ_LOG(DBG, ##__VA_ARGS__)
 
@@ -223,6 +239,29 @@ namespace kj {
   for (::kj::_::Debug::Fault f(__FILE__, __LINE__, \
            errorNumber, code, #__VA_ARGS__, ##__VA_ARGS__);; f.fatal())
 
+#if _WIN32 || __CYGWIN__
+
+#define KJ_WIN32(call, ...) \
+  if (auto _kjWin32Result = ::kj::_::Debug::win32Call(call)) {} else \
+    for (::kj::_::Debug::Fault f(__FILE__, __LINE__, \
+             _kjWin32Result, #call, #__VA_ARGS__, ##__VA_ARGS__);; f.fatal())
+// Invoke a Win32 syscall that returns either BOOL or HANDLE, and throw an exception if it fails.
+
+#define KJ_WINSOCK(call, ...) \
+  if (auto _kjWin32Result = ::kj::_::Debug::winsockCall(call)) {} else \
+    for (::kj::_::Debug::Fault f(__FILE__, __LINE__, \
+             _kjWin32Result, #call, #__VA_ARGS__, ##__VA_ARGS__);; f.fatal())
+// Like KJ_WIN32 but for winsock calls which return `int` with SOCKET_ERROR indicating failure.
+//
+// Unfortunately, it's impossible to distinguish these from BOOL-returning Win32 calls by type,
+// since BOOL is in fact an alias for `int`. :(
+
+#define KJ_FAIL_WIN32(code, errorNumber, ...) \
+  for (::kj::_::Debug::Fault f(__FILE__, __LINE__, \
+           ::kj::_::Debug::Win32Result(errorNumber), code, #__VA_ARGS__, ##__VA_ARGS__);; f.fatal())
+
+#endif
+
 #define KJ_UNIMPLEMENTED(...) \
   for (::kj::_::Debug::Fault f(__FILE__, __LINE__, ::kj::Exception::Type::UNIMPLEMENTED, \
                                nullptr, #__VA_ARGS__, ##__VA_ARGS__);; f.fatal())
@@ -251,6 +290,48 @@ namespace kj {
 
 #endif
 
+#define KJ_SYSCALL_HANDLE_ERRORS(call) \
+  if (int _kjSyscallError = ::kj::_::Debug::syscallError([&](){return (call);}, false)) \
+    switch (int error KJ_UNUSED = _kjSyscallError)
+// Like KJ_SYSCALL, but doesn't throw. Instead, the block after the macro is a switch block on the
+// error. Additionally, the int value `error` is defined within the block. So you can do:
+//
+//     KJ_SYSCALL_HANDLE_ERRORS(foo()) {
+//       case ENOENT:
+//         handleNoSuchFile();
+//         break;
+//       case EEXIST:
+//         handleExists();
+//         break;
+//       default:
+//         KJ_FAIL_SYSCALL("foo()", error);
+//     } else {
+//       handleSuccessCase();
+//     }
+
+#if _WIN32 || __CYGWIN__
+
+#define KJ_WIN32_HANDLE_ERRORS(call) \
+  if (uint _kjWin32Error = ::kj::_::Debug::win32Call(call).number) \
+    switch (uint error KJ_UNUSED = _kjWin32Error)
+// Like KJ_WIN32, but doesn't throw. Instead, the block after the macro is a switch block on the
+// error. Additionally, the int value `error` is defined within the block. So you can do:
+//
+//     KJ_SYSCALL_HANDLE_ERRORS(foo()) {
+//       case ERROR_FILE_NOT_FOUND:
+//         handleNoSuchFile();
+//         break;
+//       case ERROR_FILE_EXISTS:
+//         handleExists();
+//         break;
+//       default:
+//         KJ_FAIL_WIN32("foo()", error);
+//     } else {
+//       handleSuccessCase();
+//     }
+
+#endif
+
 #define KJ_ASSERT KJ_REQUIRE
 #define KJ_FAIL_ASSERT KJ_FAIL_REQUIRE
 #define KJ_ASSERT_NONNULL KJ_REQUIRE_NONNULL
@@ -258,7 +339,7 @@ namespace kj {
 // That is, if the assert ever fails, it indicates that the immediate surrounding code is broken.
 
 #ifdef KJ_DEBUG
-#define KJ_DLOG LOG
+#define KJ_DLOG KJ_LOG
 #define KJ_DASSERT KJ_ASSERT
 #define KJ_DREQUIRE KJ_REQUIRE
 #else
@@ -273,45 +354,44 @@ class Debug {
 public:
   Debug() = delete;
 
-  enum class Severity {
-    INFO,      // Information describing what the code is up to, which users may request to see
-               // with a flag like `--verbose`.  Does not indicate a problem.  Not printed by
-               // default; you must call setLogLevel(INFO) to enable.
-    WARNING,   // A problem was detected but execution can continue with correct output.
-    ERROR,     // Something is wrong, but execution can continue with garbage output.
-    FATAL,     // Something went wrong, and execution cannot continue.
-    DBG        // Temporary debug logging.  See KJ_DBG.
+  typedef LogSeverity Severity;  // backwards-compatibility
 
-    // Make sure to update the stringifier if you add a new severity level.
+#if _WIN32 || __CYGWIN__
+  struct Win32Result {
+    uint number;
+    inline explicit Win32Result(uint number): number(number) {}
+    operator bool() const { return number == 0; }
   };
+#endif
 
-  static inline bool shouldLog(Severity severity) { return severity >= minSeverity; }
+  static inline bool shouldLog(LogSeverity severity) { return severity >= minSeverity; }
   // Returns whether messages of the given severity should be logged.
 
-  static inline void setLogLevel(Severity severity) { minSeverity = severity; }
+  static inline void setLogLevel(LogSeverity severity) { minSeverity = severity; }
   // Set the minimum message severity which will be logged.
   //
   // TODO(someday):  Expose publicly.
 
   template <typename... Params>
-  static void log(const char* file, int line, Severity severity, const char* macroArgs,
+  static void log(const char* file, int line, LogSeverity severity, const char* macroArgs,
                   Params&&... params);
 
   class Fault {
   public:
-    template <typename... Params>
-    Fault(const char* file, int line, Exception::Type type,
-          const char* condition, const char* macroArgs, Params&&... params);
-    template <typename... Params>
-    Fault(const char* file, int line, int osErrorNumber,
+    template <typename Code, typename... Params>
+    Fault(const char* file, int line, Code code,
           const char* condition, const char* macroArgs, Params&&... params);
     Fault(const char* file, int line, Exception::Type type,
           const char* condition, const char* macroArgs);
     Fault(const char* file, int line, int osErrorNumber,
           const char* condition, const char* macroArgs);
+#if _WIN32 || __CYGWIN__
+    Fault(const char* file, int line, Win32Result osErrorNumber,
+          const char* condition, const char* macroArgs);
+#endif
     ~Fault() noexcept(false);
 
-    KJ_NORETURN(void fatal());
+    KJ_NOINLINE KJ_NORETURN(void fatal());
     // Throw the exception.
 
   private:
@@ -319,6 +399,10 @@ public:
               const char* condition, const char* macroArgs, ArrayPtr<String> argValues);
     void init(const char* file, int line, int osErrorNumber,
               const char* condition, const char* macroArgs, ArrayPtr<String> argValues);
+#if _WIN32 || __CYGWIN__
+    void init(const char* file, int line, Win32Result osErrorNumber,
+              const char* condition, const char* macroArgs, ArrayPtr<String> argValues);
+#endif
 
     Exception* exception;
   };
@@ -335,6 +419,15 @@ public:
 
   template <typename Call>
   static SyscallResult syscall(Call&& call, bool nonblocking);
+  template <typename Call>
+  static int syscallError(Call&& call, bool nonblocking);
+
+#if _WIN32 || __CYGWIN__
+  static Win32Result win32Call(int boolean);
+  static Win32Result win32Call(void* handle);
+  static Win32Result winsockCall(int result);
+  static uint getWin32ErrorCode();
+#endif
 
   class Context: public ExceptionCallback {
   public:
@@ -355,7 +448,8 @@ public:
 
     virtual void onRecoverableException(Exception&& exception) override;
     virtual void onFatalException(Exception&& exception) override;
-    virtual void logMessage(const char* file, int line, int contextDepth, String&& text) override;
+    virtual void logMessage(LogSeverity severity, const char* file, int line, int contextDepth,
+                            String&& text) override;
 
   private:
     bool logged;
@@ -381,9 +475,9 @@ public:
   static String makeDescription(const char* macroArgs, Params&&... params);
 
 private:
-  static Severity minSeverity;
+  static LogSeverity minSeverity;
 
-  static void logInternal(const char* file, int line, Severity severity, const char* macroArgs,
+  static void logInternal(const char* file, int line, LogSeverity severity, const char* macroArgs,
                           ArrayPtr<String> argValues);
   static String makeDescriptionInternal(const char* macroArgs, ArrayPtr<String> argValues);
 
@@ -391,35 +485,24 @@ private:
   // Get the error code of the last error (e.g. from errno).  Returns -1 on EINTR.
 };
 
-ArrayPtr<const char> KJ_STRINGIFY(Debug::Severity severity);
-
 template <typename... Params>
-void Debug::log(const char* file, int line, Severity severity, const char* macroArgs,
+void Debug::log(const char* file, int line, LogSeverity severity, const char* macroArgs,
                 Params&&... params) {
   String argValues[sizeof...(Params)] = {str(params)...};
   logInternal(file, line, severity, macroArgs, arrayPtr(argValues, sizeof...(Params)));
 }
 
 template <>
-inline void Debug::log<>(const char* file, int line, Severity severity, const char* macroArgs) {
+inline void Debug::log<>(const char* file, int line, LogSeverity severity, const char* macroArgs) {
   logInternal(file, line, severity, macroArgs, nullptr);
 }
 
-template <typename... Params>
-Debug::Fault::Fault(const char* file, int line, Exception::Type type,
+template <typename Code, typename... Params>
+Debug::Fault::Fault(const char* file, int line, Code code,
                     const char* condition, const char* macroArgs, Params&&... params)
     : exception(nullptr) {
   String argValues[sizeof...(Params)] = {str(params)...};
-  init(file, line, type, condition, macroArgs,
-       arrayPtr(argValues, sizeof...(Params)));
-}
-
-template <typename... Params>
-Debug::Fault::Fault(const char* file, int line, int osErrorNumber,
-                    const char* condition, const char* macroArgs, Params&&... params)
-    : exception(nullptr) {
-  String argValues[sizeof...(Params)] = {str(params)...};
-  init(file, line, osErrorNumber, condition, macroArgs,
+  init(file, line, code, condition, macroArgs,
        arrayPtr(argValues, sizeof...(Params)));
 }
 
@@ -435,6 +518,26 @@ inline Debug::Fault::Fault(const char* file, int line, kj::Exception::Type type,
   init(file, line, type, condition, macroArgs, nullptr);
 }
 
+#if _WIN32 || __CYGWIN__
+inline Debug::Fault::Fault(const char* file, int line, Win32Result osErrorNumber,
+                           const char* condition, const char* macroArgs)
+    : exception(nullptr) {
+  init(file, line, osErrorNumber, condition, macroArgs, nullptr);
+}
+
+inline Debug::Win32Result Debug::win32Call(int boolean) {
+  return boolean ? Win32Result(0) : Win32Result(getWin32ErrorCode());
+}
+inline Debug::Win32Result Debug::win32Call(void* handle) {
+  // Assume null and INVALID_HANDLE_VALUE mean failure.
+  return win32Call(handle != nullptr && handle != (void*)-1);
+}
+inline Debug::Win32Result Debug::winsockCall(int result) {
+  // Expect a return value of SOCKET_ERROR means failure.
+  return win32Call(result != -1);
+}
+#endif
+
 template <typename Call>
 Debug::SyscallResult Debug::syscall(Call&& call, bool nonblocking) {
   while (call() < 0) {
@@ -449,6 +552,20 @@ Debug::SyscallResult Debug::syscall(Call&& call, bool nonblocking) {
   return SyscallResult(0);
 }
 
+template <typename Call>
+int Debug::syscallError(Call&& call, bool nonblocking) {
+  while (call() < 0) {
+    int errorNum = getOsErrorNumber(nonblocking);
+    // getOsErrorNumber() returns -1 to indicate EINTR.
+    // Also, if nonblocking is true, then it returns 0 on EAGAIN, which will then be treated as a
+    // non-error.
+    if (errorNum != -1) {
+      return errorNum;
+    }
+  }
+  return 0;
+}
+
 template <typename... Params>
 String Debug::makeDescription(const char* macroArgs, Params&&... params) {
   String argValues[sizeof...(Params)] = {str(params)...};
@@ -463,4 +580,4 @@ inline String Debug::makeDescription<>(const char* macroArgs) {
 }  // namespace _ (private)
 }  // namespace kj
 
-#endif  // KJ_DEBUG_H_
+KJ_END_HEADER
