@@ -33,18 +33,18 @@
 // will have updated the input cursor to point to the position just past the end of what was parsed.
 // On failure, the cursor position is unspecified.
 
-#ifndef KJ_PARSE_COMMON_H_
-#define KJ_PARSE_COMMON_H_
-
-#if defined(__GNUC__) && !KJ_HEADER_WARNINGS
-#pragma GCC system_header
-#endif
+#pragma once
 
 #include "../common.h"
 #include "../memory.h"
 #include "../array.h"
 #include "../tuple.h"
 #include "../vector.h"
+#if _MSC_VER && !__clang__
+#include <type_traits>  // result_of_t
+#endif
+
+KJ_BEGIN_HEADER
 
 namespace kj {
 namespace parse {
@@ -100,7 +100,15 @@ private:
 template <typename T> struct OutputType_;
 template <typename T> struct OutputType_<Maybe<T>> { typedef T Type; };
 template <typename Parser, typename Input>
-using OutputType = typename OutputType_<decltype(instance<Parser&>()(instance<Input&>()))>::Type;
+using OutputType = typename OutputType_<
+#if _MSC_VER && !__clang__
+    std::result_of_t<Parser(Input)>
+    // The instance<T&>() based version below results in:
+    //   C2064: term does not evaluate to a function taking 1 arguments
+#else
+    decltype(instance<Parser&>()(instance<Input&>()))
+#endif
+    >::Type;
 // Synonym for the output type of a parser, given the parser type and the input type.
 
 // =======================================================================================
@@ -122,12 +130,12 @@ public:
   template <typename Other>
   constexpr ParserRef(Other&& other)
       : parser(&other), wrapper(&WrapperImplInstance<Decay<Other>>::instance) {
-    static_assert(kj::isReference<Other>(), "ParseRef should not be assigned to a temporary.");
+    static_assert(kj::isReference<Other>(), "ParserRef should not be assigned to a temporary.");
   }
 
   template <typename Other>
   inline ParserRef& operator=(Other&& other) {
-    static_assert(kj::isReference<Other>(), "ParseRef should not be assigned to a temporary.");
+    static_assert(kj::isReference<Other>(), "ParserRef should not be assigned to a temporary.");
     parser = &other;
     wrapper = &WrapperImplInstance<Decay<Other>>::instance;
     return *this;
@@ -151,7 +159,13 @@ private:
   };
   template <typename ParserImpl>
   struct WrapperImplInstance {
+#if _MSC_VER && !__clang__
+    // TODO(msvc): MSVC currently fails to initialize vtable pointers for constexpr values so
+    //   we have to make this just const instead.
+    static const WrapperImpl<ParserImpl> instance;
+#else
     static constexpr WrapperImpl<ParserImpl> instance = WrapperImpl<ParserImpl>();
+#endif
   };
 
   const void* parser;
@@ -160,8 +174,13 @@ private:
 
 template <typename Input, typename Output>
 template <typename ParserImpl>
-constexpr ParserRef<Input, Output>::WrapperImpl<ParserImpl>
+#if _MSC_VER && !__clang__
+const typename ParserRef<Input, Output>::template WrapperImpl<ParserImpl>
+ParserRef<Input, Output>::WrapperImplInstance<ParserImpl>::instance = WrapperImpl<ParserImpl>();
+#else
+constexpr typename ParserRef<Input, Output>::template WrapperImpl<ParserImpl>
 ParserRef<Input, Output>::WrapperImplInstance<ParserImpl>::instance;
+#endif
 
 template <typename Input, typename ParserImpl>
 constexpr ParserRef<Input, OutputType<ParserImpl, Input>> ref(ParserImpl& impl) {
@@ -298,25 +317,46 @@ public:
   explicit constexpr Sequence_(T&& firstSubParser, U&&... rest)
       : first(kj::fwd<T>(firstSubParser)), rest(kj::fwd<U>(rest)...) {}
 
+  // TODO(msvc): The trailing return types on `operator()` and `parseNext()` expose at least two
+  //   bugs in MSVC:
+  //
+  //     1. An ICE.
+  //     2. 'error C2672: 'operator __surrogate_func': no matching overloaded function found)',
+  //        which crops up in numerous places when trying to build the capnp command line tools.
+  //
+  //   The only workaround I found for both bugs is to omit the trailing return types and instead
+  //   rely on C++14's return type deduction.
+
   template <typename Input>
-  auto operator()(Input& input) const ->
-      Maybe<decltype(tuple(
+  auto operator()(Input& input) const
+#if !_MSC_VER || __clang__
+      -> Maybe<decltype(tuple(
           instance<OutputType<FirstSubParser, Input>>(),
-          instance<OutputType<SubParsers, Input>>()...))> {
+          instance<OutputType<SubParsers, Input>>()...))>
+#endif
+  {
     return parseNext(input);
   }
 
   template <typename Input, typename... InitialParams>
-  auto parseNext(Input& input, InitialParams&&... initialParams) const ->
-      Maybe<decltype(tuple(
+  auto parseNext(Input& input, InitialParams&&... initialParams) const
+#if !_MSC_VER || __clang__
+      -> Maybe<decltype(tuple(
           kj::fwd<InitialParams>(initialParams)...,
           instance<OutputType<FirstSubParser, Input>>(),
-          instance<OutputType<SubParsers, Input>>()...))> {
+          instance<OutputType<SubParsers, Input>>()...))>
+#endif
+  {
     KJ_IF_MAYBE(firstResult, first(input)) {
       return rest.parseNext(input, kj::fwd<InitialParams>(initialParams)...,
                             kj::mv(*firstResult));
     } else {
-      return nullptr;
+      // TODO(msvc): MSVC depends on return type deduction to compile this function, so we need to
+      //   help it deduce the right type on this code path.
+      return Maybe<decltype(tuple(
+          kj::fwd<InitialParams>(initialParams)...,
+          instance<OutputType<FirstSubParser, Input>>(),
+          instance<OutputType<SubParsers, Input>>()...))>{nullptr};
     }
   }
 
@@ -778,4 +818,4 @@ constexpr EndOfInput_ endOfInput = EndOfInput_();
 }  // namespace parse
 }  // namespace kj
 
-#endif  // KJ_PARSE_COMMON_H_
+KJ_END_HEADER
