@@ -21,14 +21,12 @@
 
 #pragma once
 
-#if defined(__GNUC__) && !KJ_HEADER_WARNINGS
-#pragma GCC system_header
-#endif
-
 #include <initializer_list>
 #include "array.h"
 #include <string.h>
 
+KJ_BEGIN_HEADER
+
 namespace kj {
   class StringPtr;
   class String;
@@ -487,6 +485,8 @@ inline _::Delimited<ArrayPtr<const T>> Stringifier::operator*(const Array<T>& ar
 //
 //    class Foo {...};
 //    inline StringPtr KJ_STRINGIFY(const Foo& foo) { return foo.name(); }
+//      // or perhaps
+//    inline String KJ_STRINGIFY(const Foo& foo) { return kj::str(foo.fld1(), ",", foo.fld2()); }
 //
 // This allows Foo to be passed to str().
 //
@@ -700,3 +700,5 @@ _::Delimited<T> delimited(T&& arr, kj::StringPtr delim) {
 constexpr kj::StringPtr operator "" _kj(const char* str, size_t n) {
   return kj::StringPtr(kj::ArrayPtr<const char>(str, n + 1));
 };
+
+KJ_END_HEADER
